# -*- coding: UTF-8 -*-
# pylint: disable=missing-docstring
"""
Provides model classes to evaluate parsed boolean tag expressions.

.. code-block:: python

    # -- Expression := a and b
    expression = And(Literal("a"), Literal("b"))
    assert True  == expression.evaluate(["a", "b"])
    assert False == expression.evaluate(["a"])
    assert False == expression.evaluate([])

    # -- Expression := a or b
    expression = Or(Literal("a"), Literal("b"))
    assert True  == expression.evaluate(["a", "b"])
    assert True  == expression.evaluate(["a"])
    assert False == expression.evaluate([])

    # -- Expression := not a
    expression = Not(Literal("a"))
    assert False == expression.evaluate(["a"])
    assert True  == expression.evaluate(["other"])
    assert True  == expression.evaluate([])

    # -- Expression := (a or b) and c
    expression = And(Or(Literal("a"), Literal("b")), Literal("c"))
    assert True  == expression.evaluate(["a", "c"])
    assert False == expression.evaluate(["c", "other"])
    assert False == expression.evaluate([])
"""


# -----------------------------------------------------------------------------
# TAG-EXPRESSION MODEL CLASSES:
# -----------------------------------------------------------------------------
class Expression(object):
    """Abstract base class for boolean expression terms of a tag-expression
    (or representing a parsed tag-expression (evaluation-tree)).
    """
    # pylint: disable=too-few-public-methods

    def evaluate(self, values):
        raise NotImplementedError()

    def __call__(self, values):
        """Call operator to make an expression-object callable."""
        return bool(self.evaluate(values))


class Literal(Expression):
    """Used as placeholder for a tag in a boolean tag-expression."""
    # pylint: disable=too-few-public-methods
    def __init__(self, name):
        super(Literal, self).__init__()
        self.name = name

    def evaluate(self, values):
        truth_value = self.name in set(values)
        return bool(truth_value)

    def __str__(self):
        return self.name

    def __repr__(self):
        return "Literal('%s')" % self.name


class And(Expression):
    """Boolean-and operation (as binary operation).

    NOTE: Class supports more than two arguments.
    """
    # pylint: disable=too-few-public-methods
    def __init__(self, *terms):
        super(And, self).__init__()
        self.terms = terms

    def evaluate(self, values):
        values_ = set(values)
        for term in self.terms:
            truth_value = term.evaluate(values_)
            if not truth_value:
                # -- SHORTCUT: Any false makes the expression false.
                return False
        # -- OTHERWISE: All terms are true.
        return True
        # -- ALTERNATIVE:
        # return all([term.evaluate(values_) for term in self.terms])

    def __str__(self):
        if not self.terms:
            return ""       # noqa
        expression_text = " and ".join([str(term) for term in self.terms])
        return "( %s )" % expression_text

    def __repr__(self):
        return "And(%s)" % ", ".join([repr(term) for term in self.terms])


class Or(Expression):
    """Boolean-or operation (as binary operation).

    NOTE: Class supports more than two arguments.
    """
    # pylint: disable=too-few-public-methods

    def __init__(self, *terms):
        super(Or, self).__init__()
        self.terms = terms

    def evaluate(self, values):
        values_ = set(values)
        for term in self.terms:
            truth_value = term.evaluate(values_)
            if truth_value:
                # -- SHORTCUT: Any true makes the expression true.
                return True
        # -- OTHERWISE: All terms are false.
        return False
        # -- ALTERNATIVE:
        # return any([term.evaluate(values_) for term in self.terms])

    def __str__(self):
        if not self.terms:
            return ""       # noqa
        expression_text = " or ".join([str(term) for term in self.terms])
        return "( %s )" % expression_text

    def __repr__(self):
        return "Or(%s)" % ", ".join([repr(term) for term in self.terms])


class Not(Expression):
    """Boolean-not operation (as unary operation)."""
    # pylint: disable=too-few-public-methods

    def __init__(self, term):
        super(Not, self).__init__()
        self.term = term

    def evaluate(self, values):
        values_ = set(values)
        return not self.term.evaluate(values_)

    def __str__(self):
        return "not ( %s )" % self.term

    def __repr__(self):
        return "Not(%r)" % self.term


class True_(Expression):    # pylint: disable=invalid-name
    """Boolean expression that is always true."""
    # pylint: disable=too-few-public-methods

    def evaluate(self, values):
        return True

    def __str__(self):
        return "true"

    def __repr__(self):
        return "True_()"



# -*- coding: UTF-8 -*-
# pylint: disable=missing-docstring
"""
Provides parsing of boolean tag expressions.

.. code-block:: python

    expression = TagExpressionParser.parse("a and (b or not c)")
    assert True == expression.evaluate(["a", "other"])
    assert "( a and ( b or not (c) ) )" == str(expression)

UNSUPPORTED:

* Support special tags w/ escaped-parens: ``@reqid\\(10\\)``
"""

from __future__ import absolute_import
from enum import Enum
from cucumber_tag_expressions.model import Literal, And, Or, Not, True_


# -----------------------------------------------------------------------------
# OPTIMIZER
# -----------------------------------------------------------------------------
# class TagExpressionOptimizer(object):
#
#     def rule_combine_and_ops(self, expression):
#         terms = getattr(expression, "terms", None)
#         if not terms:
#             return
#
#         for term in terms:
#
#     def rule_combine_or_ops(self, expression):
#         pass


# -----------------------------------------------------------------------------
# GRAMMAR DEFINITIONS:
# -----------------------------------------------------------------------------
class Associative(Enum):
    """Describes associativity of boolean operations."""
    # pylint: disable=too-few-public-methods
    LEFT = 1
    RIGHT = 2


class TokenType(Enum):
    OPERAND = 0
    OPERATOR = 1


class Token(Enum):
    """Describes tokens and their abilities for tag-expression parsing."""
    # pylint: disable=bad-whitespace
    # MAYBE: _order_ = "OR AND NOT OPEN_PARENTHESIS CLOSE_PARENTHESIS"
    OR  = ("or",  0, Associative.LEFT,  TokenType.OPERATOR)
    AND = ("and", 1, Associative.LEFT,  TokenType.OPERATOR)
    NOT = ("not", 2, Associative.RIGHT, TokenType.OPERATOR)
    OPEN_PARENTHESIS  = ("(", -2) # Java, Javascript: -2, Ruby: 1
    CLOSE_PARENTHESIS = (")", -1)

    def __init__(self, keyword, precedence, assoc=None, token_type=TokenType.OPERAND):
        self.keyword = keyword
        self.precedence = precedence
        self.assoc = assoc
        self.token_type = token_type

    @property
    def is_operation(self):
        return self.token_type is TokenType.OPERATOR

    @property
    def is_binary(self):
        return self in (Token.OR, Token.AND)

    @property
    def is_unary(self):
        return self is Token.NOT

    def has_lower_precedence_than(self, other):
        """Checks if this token has lower precedence than other token."""
        # -- pylint: disable=line-too-long
        return (
            ((self.assoc == Associative.LEFT) and (self.precedence <= other.precedence)) or
            ((self.assoc == Associative.RIGHT) and (self.precedence < other.precedence))
        )

    def matches(self, text):
        return self.keyword == text

    # def __eq__(self, other):
    #     if isinstance(other, six.string_types):
    #         # -- CONVENIENCE: token == "and"
    #         return self.matches(other)
    #     # -- OTHERWISE:
    #     return self is other


# -----------------------------------------------------------------------------
# PARSE ERRORS
# -----------------------------------------------------------------------------
class TagExpressionError(Exception):
    """Raised by parser if an invalid tag-expression is detected."""


# -----------------------------------------------------------------------------
# PARSER
# -----------------------------------------------------------------------------
class TagExpressionParser(object):
    """Parser class to parse boolean tag-expressions.
    This class uses the `Shunting Yard algorithm`_ to parse the tag-expression.

    Boolean operations:

    * and (as binary operation:  a and b)
    * or  (as binary operation:  a or b)
    * not (as unary operation:   not a)

    In addition, parenthesis can be used to group expressions, like::

        a and (b or c)
        (a and not b) or (c and d)

    EXAMPLES:

    .. code-block:: python

        # -- UNARY OPTIONS
        text11 = "not foo" = "(not foo)"
        expression = TagExpressionParser.parse(text11)
        assert False == expression.evaluate(["foo"])
        assert True  == expression.evaluate(["other"])

        # -- BINARY OPERATIONS:
        text21 = "foo and bar" = "(foo and bar)"
        expression = TagExpressionParser.parse(text21)
        assert True  == expression.evaluate(["foo", "bar"])
        assert False == expression.evaluate(["foo"])
        assert False == expression.evaluate([])

        text22 = "foo or bar"  = "(foo or bar)"
        expression = TagExpressionParser.parse(text22)
        assert True  == expression.evaluate(["foo", "bar"])
        assert True  == expression.evaluate(["foo", "other"])
        assert False == expression.evaluate([])

    .. see::

        * `Shunting Yard algorithm`_
        * http://rosettacode.org/wiki/Parsing/Shunting-yard_algorithm

    .. _`Shunting Yard algorithm`: https://en.wikipedia.org/wiki/Shunting-yard_algorithm
    """
    # pylint: disable=too-few-public-methods
    TOKEN_MAP = {token.keyword: token
                 for token in Token.__members__.values()}

    @classmethod
    def select_token(cls, text):
        """Select the token that matches the text or return None.

        :param text: Text to select the matching token.
        :return: Token object or None, if not found.
        """
        return cls.TOKEN_MAP.get(text, None)

    @classmethod
    def make_operand(cls, text):
        """Creates operand-object from parsed text."""
        # -- EXTENSION-POINT: For #406 or similar.
        return Literal(text)

    @classmethod
    def parse(cls, text):
        """Parse a textual tag-expression and return the expression (tree).

        :param text: Textual tag-expression (as string).
        :return: Tag expression (instance of :class:`Expression`).
        :raises: TagExpressionError, if the tag-expression is invalid.
        """
        # pylint: disable=too-many-branches
        # -- NOTE: Use whitespace-split to simplify tokenizing.
        #    This makes opening-/closing-parenthesis easier to parse.
        parts = text.replace("(", " ( ").replace(")", " ) ").strip().split()
        if not parts:
            #  -- CASE: Empty tag-expression is always true.
            return True_()

        def ensure_expected_token_type(token_type):
            if expected_token_type != token_type:
                message = "Syntax error. Expected %s after %s" % \
                          (expected_token_type.name.lower(), last_part)
                message = cls._make_error_description(message, parts, index)
                raise TagExpressionError(message)

        operations = []     # TOKENS: AND, OR, NOT, OPEN_PAREN, CLOSE_PAREN
        expressions = []    # Finished expressions (And, Or, Not) and Literals
        last_part = "BEGIN"
        expected_token_type = TokenType.OPERAND

        for index, part in enumerate(parts):
            token = cls.select_token(part)
            if token is None:
                # -- CASE OPERAND: Literal or ...
                ensure_expected_token_type(TokenType.OPERAND)
                expressions.append(cls.make_operand(part))
                expected_token_type = TokenType.OPERATOR
            elif token.is_unary:
                ensure_expected_token_type(TokenType.OPERAND)
                operations.append(token)
                expected_token_type = TokenType.OPERAND
            elif token.is_operation:
                ensure_expected_token_type(TokenType.OPERATOR)
                while (operations and operations[-1].is_operation and
                       token.has_lower_precedence_than(operations[-1])):
                    last_operation = operations.pop()
                    cls._push_expression(last_operation, expressions)
                operations.append(token)
                expected_token_type = TokenType.OPERAND
            elif token is Token.OPEN_PARENTHESIS:
                ensure_expected_token_type(TokenType.OPERAND)
                operations.append(token)
                expected_token_type = TokenType.OPERAND
            elif token is Token.CLOSE_PARENTHESIS:
                ensure_expected_token_type(TokenType.OPERATOR)
                while operations and operations[-1] != Token.OPEN_PARENTHESIS:
                    last_operation = operations.pop()
                    cls._push_expression(last_operation, expressions)

                if not operations:
                    # -- CASE: TOO FEW OPEN-PARENTHESIS
                    message = "Missing '(': Too few open-parens in: %s" % text
                    message = cls._make_error_description(message, parts, index)
                    raise TagExpressionError(message)
                elif operations[-1] is Token.OPEN_PARENTHESIS:
                    operations.pop()
                    expected_token_type = TokenType.OPERATOR
            last_part = part    # BETTER DIAGNOSTICS: Remember last part.

        # -- PROCESS REMAINING OPERATIONS:
        while operations:
            last_operation = operations.pop()
            if last_operation is Token.OPEN_PARENTHESIS:
                # -- CASE: TOO MANY OPEN-PARENTHESIS
                message = "Unclosed '(': Too many open-parens in: %s" % text
                raise TagExpressionError(message)
            cls._push_expression(last_operation, expressions)

        # -- FINALLY: Return boolean tag-expression.
        assert len(expressions) == 1
        expression = expressions.pop()
        return expression


    @staticmethod
    def _push_expression(token, expressions):
        """Push a new boolean-expression on the expression-stack.

        Retrieves operands for operation from the expression-stack and
        pushes the new expression onto it.

        :param token:  Token for new expression (instance of class:`Token`).
        :param expressions: Expression stack to use (as inout param).
        :raises: TagExpressionError, if too few args are in expression-stack.
        """
        def require_argcount(number):
            # -- IMPROVED DIAGNOSTICS: When things go wrong (and where).
            if len(expressions) < number:
                message = "%s: Too few operands (expressions=%r)"
                raise TagExpressionError(message % (token.keyword, expressions))

        if token is Token.OR:
            require_argcount(2)
            term2 = expressions.pop()
            term1 = expressions.pop()
            expressions.append(Or(term1, term2))
        elif token is Token.AND:
            require_argcount(2)
            term2 = expressions.pop()
            term1 = expressions.pop()
            expressions.append(And(term1, term2))
        elif token is Token.NOT:
            require_argcount(1)
            term = expressions.pop()
            expressions.append(Not(term))
        else:
            raise TagExpressionError("Unexpected token: %r" % token)    # noqa
            # expressions.append(Literal(token))

    @staticmethod
    def _make_error_description(message, parts, error_index):
        if error_index > len(parts):
            error_index = len(parts)    # noqa
        good_text_size = len(" ".join(parts[:error_index]))
        error_pos = len("Expression: ") + good_text_size + 1
        template = "Expression: {expression}\n%s^ (HERE)" % ("_" * error_pos)
        if message:                     # noqa
            template = "{message}\n" + template
        expression = " ".join(parts)
        return template.format(message=message, expression=expression)


# ----------------------------------------------------------------------------
# CONVENIENCE FUNCTIONS:
# ----------------------------------------------------------------------------
def parse(text):
    """Parse a tag-expression as text and return the expression tree.

    .. code-block:: python

        tags = ["foo", "bar"]
        tag_expression = parse("foo and bar or not baz")
        assert tag_expression.evaluate(tags) == True

    :param text:    Tag expression as text to parse.
    :param parser_class:  Optional p
    :return: Parsed expression
    """
    return TagExpressionParser.parse(text)



# -*- coding: UTF-8 -*-
"""
Python implementation of `cucumber tag-expressions`_.

Tag-expressions are used in cucumber, behave and other BDD frameworks
to select features, scenarios, etc. in `Gherkin`_ files.
Theses selected items are normally included in a test run.

.. see also:: https://cucumber.io/docs/cucumber/api/#tag-expressions

.. _`cucumber tag-expressions`: https://cucumber.io/docs/cucumber/api/#tag-expressions
.. _`Gherkin`: https://cucumber.io/docs/gherkin/reference/
.. _`behave: Gherkin`: https://behave.readthedocs.io/en/latest/philosophy.html#the-gherkin-language
"""

from __future__ import absolute_import
from .parser import parse, TagExpressionParser, TagExpressionError

__version__ = "4.0.2"
